@@ -30,7 +30,7 @@ use vars qw(
     @EXPORT
 );
 
-$VERSION = "0.80";
+$VERSION = "0.90";
 
 @EXPORT = qw(
     MainLoop
@@ -47,10 +47,28 @@ $Curses::UI::screen_too_small = 0;
 $Curses::UI::initialized      = 0;
 $Curses::UI::color_support    = 0;
 $Curses::UI::color_object     = 0;
+$Curses::UI::ncurses_mouse    = 0;
+$Curses::UI::gpm_mouse        = 0;
 
+# Detect if we should use the new moushandler
+if ($ENV{"TERM"} ne "xterm") {
+    eval { 
+	require Curses::UI::Mousehandler::GPM;
+	import Curses::UI::Mousehandler::GPM;
+    };
+    if (!$@) {
+	$Curses::UI::gpm_mouse = gpm_enable();
+	print STDERR "DEBUG: gpm_mouse: " . $Curses::UI::gpm_mouse . "\n"
+	    if $Curses::UI::debug;
+    }
+} else {
 # Detect ncurses functionality. Magic for Solaris 8
-eval { $Curses::UI::ncurses_mouse    = (Curses->can('NCURSES_MOUSE_VERSION') and
-                                 (NCURSES_MOUSE_VERSION() >= 1 ) ) };
+    eval { $Curses::UI::ncurses_mouse    = (Curses->can('NCURSES_MOUSE_VERSION') and
+					    (NCURSES_MOUSE_VERSION() >= 1 ) ) };
+    print STDERR "DEBUG: Detected mouse support value is $Curses::UI::ncurses_mouse\n" if
+	$Curses::UI::debug;
+}
+ 
 
 # ----------------------------------------------------------------------
 # Constructor
@@ -70,6 +88,7 @@ sub new()
 	-debug         => undef, # Turn on debugging mode?
 	-language      => undef, # Which language to use?
 	-mouse_support => 1,     # Do we want mouse support
+	-overlapping   => 1,     # Whether overlapping widgets are supported
 	-color_support => 0,
 	-default_colors=> 1, 
         #user data
@@ -77,7 +96,9 @@ sub new()
  
         %userargs,
 
-	-read_timeout  => -1,    # full blocking read by default
+	-read_timeout   => -1,    # full blocking read by default
+	-scheduled_code => [],
+	-added_code     => {},
     );
 
     $Curses::UI::debug = $args{-debug} 
@@ -86,6 +107,14 @@ sub new()
     $Curses::UI::ncurses_mouse = $args{-mouse_support}
         if defined $args{-mouse_support};
 
+    if ($Curses::UI::gpm_mouse) {
+	$Curses::UI::ncurses_mouse = 1;
+	$args{-read_timeout} = 0.25;
+    }
+
+    $Curses::UI::gpm_mouse = $args{-mouse_support}
+        if defined $args{-mouse_support};
+
     $Curses::UI::rootobject->fatalerror(
         "You can only initiate one Curses::UI rootobject!\n"
     ) if defined $Curses::UI::rootobject;
@@ -114,8 +143,11 @@ sub new()
 DESTROY 
 { 
     my $this = shift;
+    my $scr = $this->{-canvasscr};
+    $scr->delwin() if (defined($scr));
     endwin();
     $Curses::UI::rootobject = undef;
+    $Curses::UI::initialized = 0;
 
     if ($this->{-clear_on_exit})
     {
@@ -135,6 +167,7 @@ sub compat(;$)        { shift()->accessor('-compat',          shift()) }
 sub clear_on_exit(;$) { shift()->accessor('-clear_on_exit',   shift()) }
 sub cursor_mode(;$)   { shift()->accessor('-cursor_mode',     shift()) }
 sub lang(;$)          { shift()->accessor('-language_object', shift()) }
+sub overlapping(;$)   { shift()->accessor('-overlapping',     shift()) }
 
 # TODO: document
 sub debug(;$)         
@@ -171,15 +204,20 @@ sub layout()
     }
 
     # Mouse events if possible
-    my $old = 0;
-    if ( $Curses::UI::ncurses_mouse ) 
-    {
-	print STDERR "DEBUG: ncurses mouse events are enabled\n"
-	    if $Curses::UI::debug;
-        # In case of gpm, mousemask fails. 
-	eval { mousemask( ALL_MOUSE_EVENTS(), $old ) };
-        print STDERR "mousemask() failed: $@\n" if $@;
-    }   
+#    my $old = 0;
+#    my $mmreturn;
+#    if ( $Curses::UI::ncurses_mouse ) 
+#    {
+#	print STDERR "DEBUG: ncurses mouse events are enabled\n"
+#	    if $Curses::UI::debug;
+#        # In case of gpm, mousemask fails. (MT: Not for me, maybe GPM changed?)
+#	eval { $mmreturn = mousemask( ALL_MOUSE_EVENTS(), $old ) };
+#	if ($Curses::UI::debug) {
+#	    print STDERR "DEBUG: mousemak returned $mmreturn\n";
+#	    print STDERR "DEBUG: Old is now $old\n";
+#	    print STDERR "DEBUG: mousemask() failed: $@\n" if $@; 
+#	}
+#    }   
 
     # find the terminal size.
     my ($cols,$lines) = GetTerminalSize;
@@ -209,6 +247,37 @@ sub layout()
     return $this;    
 }
 
+sub layout_new()
+{
+    my $this = shift;
+
+    $Curses::UI::screen_too_small = 0;
+
+    # find the terminal size.
+    my ($cols,$lines) = GetTerminalSize;
+    $ENV{COLS}  = $cols;
+    $ENV{LINES} = $lines;
+
+    # Let this object present itself as a standard 
+    # Curses::UI widget, regarding size, location and
+    # drawing area. This will make it possible for
+    # child windows / widgets to layout and draw themselves.
+    #    
+    $this->{-width}  = $this->{-w} = $this->{-bw} = $cols;
+    $this->{-height} = $this->{-h} = $this->{-bh} = $lines;
+    $this->{-x}      = $this->{-y} = 0;
+#    $this->{-canvasscr} = $root;
+
+    # Walk through all contained objects and let them 
+    # layout themselves.
+    $this->layout_contained_objects;
+    
+    $Curses::UI::initialized = 1;
+    $this->draw();
+    return $this;    
+}
+
+
 # ----------------------------------------------------------------------
 # Event handling
 # ----------------------------------------------------------------------
@@ -246,6 +315,12 @@ sub do_one_event(;$)
         
     eval {curs_set($this->{-cursor_mode})};
 
+    # gpm mouse?
+    if ($Curses::UI::gpm_mouse) {
+	$this->handle_gpm_mouse_event($object);
+	doupdate();
+    }
+
     # Read a key or use the feeded key.
     my $key = $this->{-feedkey};
     unless (defined $key) {
@@ -254,13 +329,22 @@ sub do_one_event(;$)
     $this->{-feedkey} = undef;
 
     # ncurses sends KEY_RESIZE() key on resize. Ignore this key.
+    # TODO: Try to redraw and layout everything anew
+    # KEY_RESIZE doesn't seem to work right;
     if (Curses->can("KEY_RESIZE")) {
         $key = '-1' if $key eq KEY_RESIZE();
     }
+    my ($cols,$lines) = GetTerminalSize;
+    if ( ($ENV{COLS} != $cols) || ( $ENV{LINES} != $lines )) {
+	$this->layout_new();
+	$this->draw;
+    }
 
     # ncurses sends KEY_MOUSE()
     if ($Curses::UI::ncurses_mouse) {
 	if ($key eq KEY_MOUSE()) {
+	    print STDERR "DEBUG: Got a KEY_MOUSE(), handeling it\n"
+		if $Curses::UI::debug;
 	    $this->handle_mouse_event($object);
 	    doupdate();
 	    return $this;
@@ -284,6 +368,17 @@ sub do_one_event(;$)
     # Execute timer code.
     $this->do_timer;
 
+    # Execute one scheduled event;
+    if (@{$this->{-scheduled_code}}) {
+	my $code = shift @{$this->{-scheduled_code}};
+	$code->($this);
+    }
+
+    # Execute added code
+    foreach my $code (keys %{$this->{-added_code}}) {
+	$code->($this);
+    }
+
     # See if there are pending keys on input. If I do not
     # feed them to the application in this way, the screen
     # hangs in case I do a lot of input on my Solaris
@@ -297,6 +392,51 @@ sub do_one_event(;$)
     return $this;
 }
 
+# TODO: document
+sub schedule_event()
+{
+    my $this = shift;
+    my $code = shift;
+
+    $this->fatalerror(
+        "schedule_event(): callback is no CODE reference"
+    ) unless defined $code and ref $code eq 'CODE';
+
+    push @{$this->{-scheduled_code}}, $code;
+}
+
+# TODO: document
+sub add_callback()
+{
+    my $this = shift;
+    my $id   = shift;
+    my $code = shift;
+
+    $this->fatalerror(
+        "add_callback(): is is not set"
+    ) unless defined $id;
+
+    $this->fatalerror(
+        "add_callback(): callback is no CODE reference"
+    ) unless defined $code and ref $code eq 'CODE';
+
+    $this->{-added_code}->{$id} = $code;
+}
+
+# TODO: document
+sub delete_callback()
+{
+    my $this = shift;
+    my $id   = shift;
+
+    $this->fatalerror(
+      "delete_callback(): id is not set"
+    ) unless defined $id;
+
+    delete $this->{-added_code}->{$id} if
+	defined $this->{-added_code}->{$id};
+}
+
 sub draw()
 {
     my $this = shift;
@@ -498,7 +638,7 @@ sub handle_mouse_event()
 	-id     => $id,
 	-x      => $x,
 	-y      => $y,
-	-bstate => $bstate
+        -bstate => $bstate,
     );
 
     # Get the objects at the mouse event position.
@@ -509,11 +649,46 @@ sub handle_mouse_event()
     {
 	# Send the mouse-event to the object. 
 	# Leave the loop if the object handled the event.
+	print STDERR "Asking $object to handle $MEVENT{-bstate} ...\n";
 	my $return = $object->event_mouse(\%MEVENT);
 	last if defined $return and $return ne 'DELEGATE';
     }
 }
 
+sub handle_gpm_mouse_event()
+{
+    my $this = shift;
+    my $object = shift;
+    $object = $this unless defined $object;
+
+    my $MEVENT = gpm_get_mouse_event();
+    # $MEVENT from C:UI:MH:GPM is identical.
+
+    return unless $MEVENT;
+
+    my ($id, $x, $y, $z, $bstate) = unpack("sx2i3l", $MEVENT);
+    my %MEVENT = (
+	-id     => $id,
+	-x      => $x,
+	-y      => $y,
+	-bstate => $bstate,
+    );
+
+    # Get the objects at the mouse event position.
+    my $tree = $this->object_at_xy($object, $MEVENT{-x}, $MEVENT{-y});
+   
+    # Walk through the object tree, top object first.
+    foreach my $object (reverse @$tree)
+    {
+	# Send the mouse-event to the object. 
+	# Leave the loop if the object handled the event.
+
+	my $return = $object->event_mouse(\%MEVENT);
+	last if defined $return and $return ne 'DELEGATE';
+    }
+}
+
+
 sub object_at_xy($$;$)
 {
     my $this = shift;
@@ -708,17 +883,8 @@ sub process_args()
 {
     my $this = shift;        
     my $ifone = shift;
-    my %new;
-
-    if (@_ == 1) { 
-	%new = ($ifone => $_[0]);
-    } else {
-	$new{$ifone} = shift @_;
-	    if ((@_ % 2) == 0) {
-		%new = (%new, @_);
-	    }
-    }
-    return %new;
+    if (@_ == 1) { @_ = ($ifone => $_[0]) }
+    return @_;
 }
 
 sub error()
@@ -757,6 +923,21 @@ sub filebrowser()
     $this->tempdialog('Dialog::Filebrowser', %args);
 }
 
+sub dirbrowser()
+{
+    my $this = shift;
+    my %args = $this->process_args('-title', @_);
+
+    # Create title
+    unless (defined $args{-title}) {
+	my $l = $this->root->lang;
+	$args{-title} = $l->get('dir_title');
+    }
+
+    # Select a file to load from.
+    $this->tempdialog('Dialog::Dirbrowser', %args);
+}
+
 sub savefilebrowser()
 {
     my $this = shift;
@@ -1032,7 +1213,7 @@ If the B<-mouse_support> option is set to a false value
 mouse support will be disabled. This is used to override
 the auto determined value and to disable mouse support.
 
-=item * B<-userdata> < SCALAR >
+=item B<-userdata> < SCALAR >
 
 This option specifies a user data that can be retrieved with
 the B<userdata>() method.  It is usefull to store application's
@@ -1094,6 +1275,21 @@ Starts a Tk like main loop that will handle input and events.
 
 Same as B<mainloop>, for Tk compatibility.
 
+=item B<schedule_event> ( CODE )
+
+The schedule_event method adds a method to the mainloop. This
+method is executed one time after the input handler has run and
+deleted from the mainloop afterwards.
+
+=item B<add_callback> ( ID, CODE)
+
+This method lets you add a callback into the mainloop permanently.
+The code is executed after the input handler has run.
+
+=item B<delete_callback> ( ID )
+
+This method deletes the CODE specified by ID from the mainloop.
+
 =item B<usemodule> ( CLASSNAME )
 
 Loads the with CLASSNAME given module.
@@ -1139,8 +1335,8 @@ L<Curses::UI::Dialog::Basic|Curses::UI::Dialog::Basic>.
 Example:
 
     my $yes = $cui->dialog(
-        -message => "Hello, world?");
-        -buttons => ['< Yes >','< No >']
+        -message => "Hello, world?",
+        -buttons => ['< Yes >','< No >'],
         -values  => [1,0],
         -title   => 'Question',
     );
@@ -1275,170 +1471,9 @@ can be used to fast change all colors in a Curses::UI application.
 L<Curses>
 L<Curses::UI::Container|Curse::UI::Container|Curses::UI::Color>,
 
-
-
-
 =head1 BASIC TUTORIAL
 
-=head1 First requirements
-
-Any perl program that uses Curses::UI needs to include
-"use Curses::UI". A program should also use "use strict"
-and the B<-w> switch to ensure the program is working
-without common errors (but of course Curses::UI will work
-without them). After that an instance of Curses::UI must
-be created. From now on, this instance will be called
-"the UI".
-
-    #!/usr/bin/perl -w
-
-    use strict;
-    use Curses::UI;
-    my $cui = new Curses::UI;
-
-=head1 Create windows
-
-After the initialization has been done, windows can be
-added to the UI. You will always have to do this. It is not
-possible to add widgets to the UI directly. Here is an
-example that creates a window with a title and a border,
-which has a padding of 2 (For the explanation of $cui->add, 
-see the Curses::UI::Container manual page).
-
-    my $win1 = $cui->add(
-        'win1', 'Window',
-        -border => 1,
-        -title => 'My first Curses::UI window!',
-        -pad => 2,
-    );
-
-Well... that's fun! Let's add another window! And let's
-give it more padding, so that the window is smaller than
-the previous one.
-    
-    my $win2 = $cui->add(
-        'win2', 'Window',
-        -border => 1,
-        -title => 'My second Curses::UI window!',
-        -pad => 6,
-    );
-
-=head1 Add some widgets
-
-Now that we have a couple of windows, we can add widgets
-to them. We'll add a popupmenu and some buttons to 
-the first window.
-
-    my $popup = $win1->add(
-        'popup', 'Popupmenu',
-        -x => 2,
-        -y => 2,
-        -sbborder => 1, 
-        -values => [ 1, 2, 3, 4, 5 ],
-        -labels => {
-            1 => 'One', 
-            2 => 'Two', 
-            3 => 'Three', 
-            4 => 'Four', 
-            5 => 'Five', 
-        },
-    );
-
-    my $but1 = $win1->add(
-        'buttons', 'Buttonbox',
-        -x => 2,
-        -y => 4,
-        -buttons => [
-            { -label => '< goto window 2 >', -value => 'goto2' },
-            { -label => '< Quit >', -value => 'quit' },
-        ],
-    ); 
-
-We'll add a texteditor and some buttons to the second window. Look how
-we can use padding to do some basic layouting. We do not specifiy the
-with and height of the TextEditor widget, so the widget will stretch
-out itself as far as possible. But because of the -padbottom option,
-it will leave some space to put the buttons. By specifying a negative
--y offset for the buttons, we have a scalable application. Resize the
-screen and the widgets will follow!
-    
-    my $editor = $win2->add(
-        'editor', 'TextEditor',
-        -border => 1,
-        -vscrollbar => 1,
-        -wrapping => 1,
-        -x => 2,
-        -y => 1,
-        -padright => 2,
-        -padbottom => 3,
-    );
-
-    my $but2 = $win2->add(
-        'buttons', 'Buttonbox',
-        -x => 2,
-        -y => -2,
-        -buttons => [
-            { -label => '< goto window 1 >', -value => 'goto1' },
-            { -label => '< Quit >', -value => 'quit' },
-        ],
-    );
-
-=head1 Specify when the windows will loose their focus
-
-We have a couple of buttons on each window. As soon as a 
-button is pressed, it will have the window loose its
-focus (buttons will have any kind of Container object
-loose its focus). You will only have to do something
-if this is not the desired behaviour. 
-
-If you want the buttons themselves to loose focus if 
-pressed, then change its routine for the "return" 
-binding from "LEAVE_CONTAINER" to "LOOSE_FOCUS". Example:
-
-    $but1->set_routine('loose-focus', 'LOOSE_FOCUS');
-
-To make things a bit more snappy, we want to add some
-shortcut keys to the appliction:
-
-    CTRL+Q : Quit the application
-    CTRL+N : Go to the next window
-
-This can be done by assigning "returnkeys" to a window.
-Each widget in the window will get extra keybindings to
-have the window loose its focus if one of the returnkeys
-is pressed. For our application we can set the desired
-shortcut keys like this:
-
-    $win1->returnkeys("\cN", "\cQ");
-    $win2->returnkeys("\cN", "\cQ");
-
-From now on both windows will loose focus if either CTRL+Q
-or CTRL+N is pressed. Important: make sure that returnkeys 
-are assigned to a window _after_ all windgets have been 
-added. 
-
-=head1 The main loop
-
-The main loop is called by
-
-    $cui->MainLoop;
-
-=head1 Add a good-bye dialog
-
-Curses::UI has a couple of methods that are easy for showing dialogs 
-(see the METHODS section below). We'll use a dialogbox to say goodbye
-to the user of our program. After the mainloop we add:
-
-    $cui->dialog("Bye bye!");
-
-=head1 You're done!
-
-We have built a genuine Curses::UI application! Not that it is a
-very useful one, but who cares? Now try out if it works like you 
-think it should. The complete source code of this application can 
-be found in the examples directory of the distribution 
-(examples/demo-Curses::UI).
-
+see 'perldoc Curses::UI::Tutorial'
 
 =head1 REFERENCES
 
